<?php

    class KGatewayNetworkException extends Exception
    {
    }

    class KGatewayNetwork
    {
        /**
         * array Configuration file lines as array
         */
        protected $data = array( );
        /**
         * array Network interfaces
         */
        protected $interfaces = array( );
        /**
         * string File name
         */
        protected $file = null;
        /**
         * int Input/Output error
         */
        const EIO = 1;
        /**
         * int Runtime error
         */
        const ERUNTIME = 2;
        /**
         * int Invalid parameter/value
         */
        const EINVALID = 3;
        /**
         * int Permission denied
         */
        const ENOPERM = 4;
        /**
         * array Supported methods
         */
        static public $methods = array( 'inet' => array( 'dhcp', 'static' ) );
        // @formatter:off
        /**
         * array Stanzas from interface(5)
         * 
         * Format: __family__ => array
         *           __method__ => array
         *             valid stanza => array
         *               verification method => string|array
         * 
         * The verification method can be a string or array. All verification
         * methods accept at least one value (the value to be verified) and
         * return true or false. Optionaly the methods can receive a second
         * value that should be supplied defining this verification method
         * as array and the second position should be the value to be supplied
         * as second parameter to the 'real' verification method.
         */
        static protected $stanzas = array(
            'inet' => array(
                'loopback' => array(),
                'static' => array(
                    'address' => array( 'checkIpv4', array( 'check-ebs-subnet' => true ) ),
                    'netmask' => 'checkIpv4',
                    'broadcast' => array( 'checkIpv4', array( 'empty' => true ) ),
                    'network' => array( 'checkIpv4', array( 'empty' => true ) ),
                    'metric' => array( 'checkInteger', array( 'empty' => true ) ),
                    'gateway' => array( 'checkIpv4', array( 'check-ebs-subnet' => true, 'empty' => true ) ),
                    'pointopoint' => array( 'checkIpv4', true ),
                    'media' => null,
                    'hwaddress' => array( 'checkMacAddress', array( 'emtpy' => true ) ),
                    'mtu' => array( 'checkInteger', array( 'min' => 64, 'max' => 1500, 'emtpy' => true ) )
                ),
                'dhcp' => array(
                    'hostname' => 'checkFqdn',
                    'leasehours' => 'checkInteger',
                    'leasetime' => 'checkInteger',
                    'vendor' => null,
                    'client' => 'checkAlwaysInvalid',
                    'hwaddress' => 'checkMacAddress'
                ),
                'bootp' => array(
                    'bootfile' => null,
                    'server' => 'checkIpv4OrIpv6',
                    'hwaddr' => 'checkMacAddress'
                ),
                'ppp' => array(
                    'provider' => null
                ),
                'wvdial' => array(
                    'provider' => null
                )
            ),
            'ipx' => array(
                'static' => array(
                    'frame' => 'checkIpxFrame',
                    'netnum' => 'checkIpxNetNum'
                ),
                'dynamic' => array(
                    'frame' => 'checkIpxFrame'
                )
            ),
            'inet6' => array(
                'loopback' => array(),
                'static' => array(
                    'address' => 'checkIpv6',
                    'netmask' => 'checkIpv6Mask',
                    'gateway' => array( 'checkIpv6', array( 'empty' => true ) ),
                    'media' => null,
                    'hwaddress' => array( 'checkMacAddress', array( 'empty' => true ) ),
                    'mtu' => array( 'checkInteger', array( 'min' => 64, 'max' => 1500, 'empty' => true ) )
                ),
                'v4tunnel' => array(
                    'address' => 'checkIpv6',
                    'netmask' => 'checkIpv6Mask',
                    'endpoint' => 'checkIpv4',
                    'local' => 'checkIpv4',
                    'gateway' => 'checkIpv6',
                    'ttl' => array( 'checkInteger', array( 'min' => 1, 'max' => 255 ) )
                )
            )
        );
        /**
         * array Stanzas from resolvconf(8)
         * 
         * Format: stanza => array
         *           verification method => string|array
         */
        static protected $stanzasResolvconf = array(
            'dns-nameservers' => array( 'checkIpv4OrIpv6List', array( 'emtpy' => true ) ),
            'dns-search' => array( 'checkFqdn', array( 'empty' => true ) )
        );
        // @formatter:on
        /**
         * array Supported dns methods
         */
        static public $dnsmethods = array( 'dhcp', 'static' );

        public function __construct( $file )
        {
            $this->file = $file;
        }

        /**
         * Parse the file and create internal representation
         */
        public function load( )
        {
            if( !file_exists( $this->file ) )
            {
                throw new KGatewayNetworkException( 'gateway-network-interface-error-configuration-not-found', self::EIO );
            }
            $this->data = file( $this->file, FILE_IGNORE_NEW_LINES );
            if( empty( $this->data ) )
            {
                throw new KGatewayNetworkException( 'gateway-network-interface-error-configuration-read-failed', self::EIO );
            }
            $currentiface = null;
            // parse each line and generate internal representation
            foreach( $this->data as $idx => $line )
            {
                $line = trim( $line );
                if( empty( $line ) || $line[0] == '#' )
                {
                    continue;
                }
                else
                {
                    $_info = explode( ' ', $line );
                    switch( $_info[0] )
                    {
                        case 'iface':
                            $currentiface = $_info[1];
                            // @formatter:off
                            $this->interfaces[$currentiface] = array(
                                'info' => array( 'start' => $idx, 'end' => $idx, 'dirty' => false ),
                                'properties' => array( 'family' => $_info[2], 'method' => $_info[3] ),
                                'config' => array( ),
                                'immutable' => array( )
                            );
                            // @formatter:on
                            break;
                        // resolvconf(8) stanzas
                        case 'dns-nameservers':
                        case 'dns-search':
                            $field = $_info[0];
                            unset( $_info[0] );
                            $this->interfaces[$currentiface]['config'][$field] = implode( ' ', $_info );
                            $this->interfaces[$currentiface]['info']['end'] = $idx;
                            break;
                        case 'pre-up':
                        case 'up':
                        case 'post-up':
                        case 'pre-down':
                        case 'down':
                        case 'post-down':
                            $field = $_info[0];
                            unset( $_info[0] );
                            $this->interfaces[$currentiface]['immutable'][$field] = implode( ' ', $_info );
                            $this->interfaces[$currentiface]['info']['end'] = $idx;
                            break;
                        // stop filling the current interface
                        case 'mapping':
                        case 'auto':
                        case 'allow-hotplug':
                        case 'allow-auto':
                            if( !is_null( $currentiface ) )
                            {
                                $currentiface = null;
                            }
                            break;
                        default:
                            if( !is_null( $currentiface ) )
                            {
                                $_family = $this->interfaces[$currentiface]['properties']['family'];
                                $_method = $this->interfaces[$currentiface]['properties']['method'];
                                $field = $_info[0];
                                if( array_key_exists( $_family, self::$stanzas ) && array_key_exists( $_method, self::$stanzas[$_family] ) && array_key_exists( $field, self::$stanzas[$_family][$_method] ) )
                                {
                                    $field = $_info[0];
                                    unset( $_info[0] );
                                    $this->interfaces[$currentiface]['config'][$field] = implode( ' ', $_info );
                                    $this->interfaces[$currentiface]['info']['end'] = $idx;
                                }
                                unset( $_family );
                                unset( $_method );
                            }
                            break;
                    }
                }
            }
        }

        /**
         * Save the configuration
         */
        public function save( )
        {
            if( count( $this->data ) == 0 || count( $this->interfaces ) == 0 )
            {
                throw new KGatewayNetworkException( 'gateway-network-interface-error-invalid-config', self::EINVALID );
            }
            $_wheretochange = array( );
            // Check for dirty interface configurations and required stanzas
            foreach( $this->interfaces as $interface => &$config )
            {
                if( $config['info']['dirty'] === false )
                {
                    // ignore interfaces that hasn't changed
                    continue;
                }
                switch( $config['properties']['family'] )
                {
                    case 'inet':
                        switch( $config['properties']['method'] )
                        {
                            case 'static':
                                if( !array_key_exists( 'address', $config['config'] ) || empty( $config['config']['address'] ) )
                                {
                                    throw new KGatewayNetworkException( 'gateway-network-interface-error-address-required', self::EINVALID );
                                }
                                elseif( !array_key_exists( 'netmask', $config['config'] ) || empty( $config['config']['netmask'] ) )
                                {
                                    throw new KGatewayNetworkException( 'gateway-network-interface-error-netmask-required', self::EINVALID );
                                }
                                break;
                            // dont validate unknown methods
                            default:
                                break;
                        }
                        break;
                    //TODO: implement support to IPv6 (inet6) and IPX as needed
                    default:
                        // for now, all unsupported families are just ignored
                        // during validation
                        break;
                }
                $_wheretochange[$config['info']['start']] = $interface;
            }

            // nothing to do, just bailout
            if( count( $_wheretochange ) == 0 )
            {
                return;
            }

            // configuration OK, update $this->data and save it
            $_data = array( );

            $skip = 0;
            foreach( $this->data as $_idx => $_line )
            {
                if( array_key_exists( $_idx, $_wheretochange ) )
                {
                    $interface = $_wheretochange[$_idx];
                    $config = &$this->interfaces[$interface];
                    $skip = $config['info']['end'] - $config['info']['start'];

                    $_data[] = "iface {$interface} {$config['properties']['family']} {$config['properties']['method']}";
                    foreach( $config['config'] as $key => $value )
                    {
                        if( !empty( $value ) )
                        {
                            $_data[] = "    {$key} {$value}";
                        }
                    }
                    foreach( $config['immutable'] as $key => $value )
                    {
                        if( !empty( $value ) )
                        {
                            $_data[] = "    {$key} {$value}";
                        }
                    }
                }
                elseif( $skip > 0 )
                {
                    // skip lines from the old configuration file
                    $skip--;
                    continue;
                }
                else
                {
                    $_data[] = $_line;
                }
            }

            $_data[] = '';
            // last line break
            $this->data = $_data;
            unset( $_data );
            $_tmpname = "{$this->file}.in";
            if( file_put_contents( $_tmpname, implode( "\n", $this->data ) ) === false )
            {
                throw new KGatewayNetworkException( 'gateway-network-interface-error-write-failed', self::EIO );
            }
            if( rename( $_tmpname, $this->file ) === false )
            {
                unlink( $_tmpname );
                throw new KGatewayNetworkException( 'gateway-network-interface-error-write-failed', self::EIO );
            }
        }

        /**
         * Get available interfaces
         */
        public function getInterfaces( )
        {
            return array_keys( $this->interfaces );
        }

        /**
         * Get interface properties
         */
        public function getProperties( $interface )
        {
            if( array_key_exists( $interface, $this->interfaces ) )
            {
                return $this->interfaces[$interface]['properties'];
            }
            else
            {
                throw new KGatewayNetworkException( 'gateway-network-interface-error-not-found', self::ERUNTIME );
            }
        }

        /**
         * Set interface properties
         */
        public function setProperties( $interface, $properties, $config )
        {
            if( $interface == 'lo' )
            {
                throw new KGatewayNetworkException( 'gateway-network-interface-error-loopback', self::ENOPERM );
            }
            elseif( !is_array( $properties ) )
            {
                throw new KGatewayNetworkException( 'gateway-network-interface-error-invalid-properties', self::EINVALID );
            }
            elseif( !is_array( $config ) )
            {
                throw new KGatewayNetworkException( 'gateway-network-interface-error-invalid-config', self::EINVALID );
            }
            elseif( array_key_exists( $interface, $this->interfaces ) )
            {
                $this->interfaces[$interface]['properties'] = $properties;
                $this->setConfiguration( $interface, $config );
            }
            else
            {
                throw new KGatewayNetworkException( 'gateway-network-interface-error-not-found', self::ERUNTIME );
            }
        }

        /**
         * Get interface configuration
         */
        public function getConfiguration( $interface )
        {
            if( array_key_exists( $interface, $this->interfaces ) )
            {
                return $this->interfaces[$interface]['config'];
            }
            else
            {
                throw new KGatewayNetworkException( 'gateway-network-interface-error-not-found', self::ERUNTIME );
            }
        }

        /**
         * Set interface configuration
         */
        public function setConfiguration( $interface, $config )
        {
            if( $interface == 'lo' )
            {
                throw new KGatewayNetworkException( 'gateway-network-interface-error-loopback', self::ENOPERM );
            }
            elseif( !is_array( $config ) )
            {
                throw new KGatewayNetworkException( 'gateway-network-interface-error-invalid-config', self::EINVALID );
            }
            elseif( array_key_exists( $interface, $this->interfaces ) )
            {
                $_family = $this->interfaces[$interface]['properties']['family'];
                $_method = $this->interfaces[$interface]['properties']['method'];
                foreach( $config as $key => &$value )
                {
                    // interfaces(5) stanzas
                    if( array_key_exists( $key, self::$stanzas[$_family][$_method] ) )
                    {
                        $this->check( $value, self::$stanzas[$_family][$_method][$key] );
                    }
                    // resolvconf(8) stanzas
                    elseif( array_key_exists( $key, self::$stanzasResolvconf ) )
                    {
                        $this->check( $value, self::$stanzasResolvconf[$key] );
                    }
                    // unknown stanza
                    else
                    {
                        throw new KGatewayNetworkException( 'gateway-network-interface-error-invalid-stanza', self::EINVALID );
                    }
                }
                $this->interfaces[$interface]['config'] = $config;
                $this->interfaces[$interface]['info']['dirty'] = true;
            }
            else
            {
                throw new KGatewayNetworkException( 'gateway-network-interface-error-not-found', self::ERUNTIME );
            }
        }

        /**
         * Get the interface real mac address
         */
        public function getMacAddress( $interface )
        {
            if( array_key_exists( $interface, $this->interfaces ) )
            {
                return trim( file_get_contents( "/sys/class/net/{$interface}/address" ) );
            }
            else
            {
                throw new KGatewayNetworkException( 'gateway-network-interface-error-not-found', self::ERUNTIME );
            }
        }

        // ********************** validation methods **************************
        /**
         * Check value using supplied configuration
         *
         * @param value string
         * @param configuration string|array|null
         *
         * @throws KGatewayNetworkException
         */
        protected function check( &$value, $configuration )
        {
            $validationMethod = null;
            $validationData = null;
            if( is_null( $configuration ) )
            {
                // no validator
                return;
            }
            elseif( is_array( $configuration ) )
            {
                $validationMethod = $configuration[0];
                $validationData = $configuration[1];
            }
            else
            {
                $validationMethod = $configuration;
            }

            if( !method_exists( $this, $validationMethod ) )
            {
                throw new KGatewayNetworkException( 'gateway-network-interface-error-validator-not-found', self::ERUNTIME );
            }
            else
            {
                if( is_null( $validationData ) )
                {
                    $ok = $this->$validationMethod( $value );
                }
                else
                {
                    $ok = $this->$validationMethod( $value, $validationData );
                }

                if( !$ok )
                {
                    throw new KGatewayNetworkException( 'gateway-network-interface-error-invalid-value', self::EINVALID );
                }
            }
        }

        /**
         * Check if value is in ipv4 format
         *
         * data array valid options:
         *   empty => boolean - if true, value can be empty
         *   check-ebs-subnet => boolean - if true, check the reserved subnet
         *
         * @param value string
         * @param data array options
         * network
         * @return boolean
         */
        public function checkIpv4( &$value, $data = array() )
        {
            if( array_key_exists( 'empty', $data ) && $data['empty'] && empty( $value ) )
            {
                return true;
            }
            // cpu <-> ebs network
            if( array_key_exists( 'check-ebs-subnet', $data ) && $data['check-ebs-subnet'] && substr( $value, 0, 10 ) == '172.31.249' )
            {
                throw new KGatewayNetworkException( 'gateway-network-interface-error-restricted-subnet', self::EINVALID );
            }
            $parts = explode( '.', $value );
            if( count( $parts ) == 4 )
            {
                foreach( $parts as $part )
                {
                    if( (int)$part < 0 || (int)$part > 255 )
                    {
                        break;
                    }
                }
                return true;
            }
            return false;
        }

        public function checkMacAddress( &$value, $data = array() )
        {
            if( array_key_exists( 'empty', $data ) && $data['empty'] && empty( $value ) )
            {
                return true;
            }
            return false;
        }

        public function checkFqdn( &$value, $data = array() )
        {
            if( array_key_exists( 'empty', $data ) && $data['empty'] && empty( $value ) )
            {
                return true;
            }
            return preg_match( '/([a-zA-Z0-9\-_]+\.)+[a-zA-Z0-9]{2,3}/', $value ) === 1;
        }

        public function checkInteger( &$value, $data = array() )
        {
            if( array_key_exists( 'empty', $data ) && $data['empty'] && empty( $value ) )
            {
                return true;
            }
            if( is_numeric( $value ) )
            {
                if( array_key_exists( 'min', $data ) || array_key_exists( 'max', $data ) )
                {
                    $min = array_key_exists( 'min', $range ) ? $range['min'] : -PHP_INT_MAX;
                    $max = array_key_exists( 'max', $range ) ? $range['max'] : PHP_INT_MAX;
                    if( (int)$value < $min || (int)$value > $max )
                    {
                        return false;
                    }
                }
                return true;
            }
            return false;
        }

        public function checkAlwaysInvalid( &$value, $data = null )
        {
            return false;
        }

        public function checkIpv4OrIpv6( &$value, $data = array() )
        {
            return $this->checkIpv4( $value, $data ) || $this->checkIpv6( $value, $data );
        }

        public function checkIpxFrame( &$value, $data = array() )
        {
            if( array_key_exists( 'empty', $data ) && $data['empty'] && empty( $value ) )
            {
                return true;
            }
            return false;
        }

        public function checkIpxNetNum( &$value, $data = array() )
        {
            if( array_key_exists( 'empty', $data ) && $data['empty'] && empty( $value ) )
            {
                return true;
            }
            return false;
        }

        public function checkIpv6( &$value, $data = array() )
        {
            if( array_key_exists( 'empty', $data ) && $data['empty'] && empty( $value ) )
            {
                return true;
            }
            return false;
        }

        public function checkIpv6Mask( &$value, $data = array() )
        {
            if( array_key_exists( 'empty', $data ) && $data['empty'] && empty( $value ) )
            {
                return true;
            }
            return false;
        }

        /**
         * Check if all elements from the list are ipv4 or ipv6 strings, also
         * convert the value to native representation, if needed
         *
         * @param array|string list values list
         * @return boolean
         */
        public function checkIpv4OrIpv6List( &$list, $data = array() )
        {
            if( array_key_exists( 'empty', $data ) && $data['empty'] && empty( $list ) )
            {
                return true;
            }
            if( is_array( $list ) )
            {
                $l = &$list;
            }
            else
            {
                $l = explode( ' ', $list );
            }

            if( count( $l ) == 0 )
            {
                return false;
            }

            foreach( $l as &$v )
            {
                if( !$this->checkIpv4OrIpv6( $v, $data ) )
                {
                    return false;
                }
            }
            if( is_array( $list ) )
            {
                $_l = implode( ' ', $list );
                $list = $_l;
            }
            return true;
        }

    }
