<?php
    class KLicenseManagerException extends Exception
    {
    }

    class KLicenseException extends Exception
    {
    }

    /**
     * Licenses managment
     */
    class KLicenseManager
    {
        /**
         * string Licenses directory
         */
        protected $licensesDir = null;
        /**
         * DateTime Current date time
         */
        protected $now = null;
        /**
         * DateTime +15 days -- used to check if license will expire soon
         */
        protected $willexpire = null;
        /**
         * int Input/Output errors
         */
        const EIO = 1;
        /**
         * int Invalid data
         */
        const EINVALID = 2;

        public function __construct( )
        {
            $app = KSystem::getInstance( );
            if( PHP_OS == 'WINNT' )
            {
                $this->licensesDir = "{$app->getInstallDir()}\\licenses";
            }
            else
            {
                $this->licensesDir = "{$app->getInstallDir()}/licenses";
            }
            if( !is_dir( $this->licensesDir ) )
            {
                $this->licensesDir = null;
                throw new KLicenseManagerException( 'license-manager-error-directory-not-found', self::EIO );
            }
            $this->now = new DateTime( );
            $this->willexpire = new DateTime( );
            // +15 days
            $parts = explode( '-', date( 'Y-m-d', time( ) + (15 * 86400) ) );
            $this->willexpire->setDate( $parts[0], $parts[1], $parts[2] );
        }

        /**
         * Scan license directory and return an array of licenses objects
         *
         * @return array of licenses
         */
        public function getLicenses( )
        {
            $licenses = array( );
            if( PHP_OS == 'WINNT' )
            {
                $files = glob( "{$this->licensesDir}\\*.klf" );
            }
            else
            {
                $files = glob( "{$this->licensesDir}/*.klf" );
            }
            if( $files === false || !is_array( $files ) )
            {
                throw new KLicenseManagerException( 'license-manager-error-directory-read-failed', self::EIO );
            }
            elseif( count( $files ) > 0 )
            {
                foreach( $files as $file )
                {
                    try
                    {
                        $licenses[] = $this->getLicense( basename( $file ) );
                    }
                    catch( KLicenseException $e )
                    {
                        /* VOID */
                    }
                }
            }
            return $licenses;
        }

        /**
         * Get license object
         *
         * @param name string License (file) name
         * @return KLicense
         */
        public function getLicense( $name )
        {
            return new KLicense( $this, $name );
        }

        /**
         * Get license as string
         *
         * @param name string License (file) name
         * @return string
         */
        public function getLicenseRaw( $name )
        {
            return file_get_contents( KLicense::getFilenameWithFullPath( $this, $name ) );
        }

        /**
         * Delete license from licenses directory
         *
         * @param name string License (file) name
         */
        public function deleteLicense( $name )
        {
            if( unlink( KLicense::getFilenameWithFullPath( $this, $name ) ) === false )
            {
                throw new KLicenseManagerException( 'license-manager-error-unlink-failed', self::EIO );
            }
        }

        /**
         * Upload license to licenses directory
         *
         * @param info array Information array (as part of $_FILE super-global)
         */
        public function uploadLicense( $info )
        {
            // .klt
            $_tmpname = substr( $info['name'], 0, strlen( $info['name'] ) - 1 ) . 't';
            $_tmpfilename = KLicense::getFilenameWithFullPath( $this, $_tmpname, false );
            if( move_uploaded_file( $info['tmp_name'], $_tmpfilename ) === false )
            {
                throw new KLicenseManagerException( 'license-manager-error-upload-failed', self::EIO );
            }

            try
            {
                $newlicense = new KLicense( $this, $_tmpname );
                $_name = "{$newlicense->id} - K{$newlicense->serial_number} - {$newlicense->getTypeAsString( ' ', '%v %n' )} - {$newlicense->customer_info}.klf";
                $_filename = KLicense::getFilenameWithFullPath( $this, $_name, false );
                $currentlicense = null;
                foreach( $this->getLicenses() as $license )
                {
                    if( $newlicense->id == $license->id )
                    {
                        $currentlicense = $license;
                        break;
                    }
                }

                // if license already exists, then check if the new has a newer
                // issue_date
                if( !is_null( $currentlicense ) )
                {
                    if( $newlicense->getIssued( ) < $currentlicense->getIssued( ) )
                    {
                        throw new KLicenseManagerException( 'license-manager-error-current-license-is-newer', self::EINVALID );
                    }
                }

                // commit new license
                if( rename( $_tmpfilename, $_filename ) === false )
                {
                    throw new KLicenseManagerException( 'license-manager-error-rename-failed', self::EIO );
                }

                $_tmpname = null;

                // if their name are different, remove the "current"
                if( !is_null( $currentlicense ) && $currentlicense->getFilename( ) != $_name )
                {
                    $this->deleteLicense( $currentlicense->getFilename( ) );
                }
            }
            // gambi finally, rah!
            catch( Exception $e )
            {
                try
                {
                    if( !is_null( $_tmpname ) )
                    {
                        $this->deleteLicense( $_tmpname );
                    }
                }
                catch( Exception $_e )
                {
                    /* ignore errors */
                }
                throw $e;
            }
        }

        /**
         * Get current time as object
         *
         * @return DateTime
         */
        public function getNow( )
        {
            return $this->now;
        }

        /**
         * Get time +15 days
         *
         * @return DateTime
         */
        public function getWillExpire( )
        {
            return $this->willexpire;
        }

        /**
         * Get licenses directory
         */
        public function getLicensesDirectory( )
        {
            return $this->licensesDir;
        }

    }

    /**
     * License object
     */
    class KLicense
    {
        // string License file name
        protected $filename = null;
        // KLicenseManager manager
        protected $manager = null;
        // array License data
        protected $license = array( );
        // int License type
        protected $type = 0;
        // DateTime Expires date object
        protected $expires = null;
        // DateTime Issued date object
        protected $issued = null;

        // int Input/Output error
        const EIO = 1;
        // int Invalid license error
        const EINVALID = 2;

        // Types from KLicense.cpp
        // int Unknown type
        const TYPE_NONE = 0;
        // int FAX_CHANNELS
        const TYPE_FAX = 1;
        // int ISUP_CIRCUITS
        const TYPE_ISUP = 2;
        // int MTP2_LINKS
        const TYPE_MTP2 = 4;
        // int MTP_DUAL_MODE
        const TYPE_MTP = 8;
        // int G729_CHANNELS
        const TYPE_G729 = 16;
        // int VOIP_CHANNELS
        const TYPE_VOIP = 32;

        public function __construct( KLicenseManager $manager, $name )
        {
            $this->manager = $manager;
            $this->filename = $name;

            // parse license
            $data = file( self::getFilenameWithFullPath( $manager, $name ), FILE_IGNORE_NEW_LINES );
            if( $data[0] != 'KHOMP LICENSE' )
            {
                $this->filename = null;
                throw new KLicenseException( 'license-error-invalid', self::EINVALID );
            }
            unset( $data[0] );
            foreach( $data as $line )
            {
                // end of license data
                if( strlen( $line ) > 0 && $line[0] == chr( 27 ) )
                {
                    break;
                }
                $line = trim( $line );
                if( strlen( $line ) > 0 )
                {
                    $keypos = strpos( $line, '=' );
                    $key = strtolower( substr( $line, 0, $keypos ) );
                    $value = substr( $line, $keypos + 1 );
                    $value = is_string( $value ) ? $value : '';
                    $this->license[$key] = $value;
                }
            }

            if( $this->hasKey( 'fax_channels' ) )
            {
                $this->type |= self::TYPE_FAX;
            }
            if( $this->hasKey( 'isup_circuits' ) )
            {
                $this->type |= self::TYPE_ISUP;
            }
            if( $this->hasKey( 'mtp2_links' ) )
            {
                $this->type |= self::TYPE_MTP2;
            }
            if( $this->hasKey( 'mtp_dual_mode' ) )
            {
                $this->type |= self::TYPE_MTP;
            }
            if( $this->hasKey( 'g729_channels' ) )
            {
                $this->type |= self::TYPE_G729;
            }
            if( $this->hasKey( 'voip_channels' ) )
            {
                $this->type |= self::TYPE_VOIP;
            }

            // easy to use issued_date and expiration_date objects
            $parts = explode( '-', $this->issued_date );
            $this->issued = new DateTime( );
            $this->issued->setDate( $parts[0], $parts[1], $parts[2] );
            $parts = explode( '-', $this->expiration_date );
            $this->expires = new DateTime( );
            $this->expires->setDate( $parts[0], $parts[1], $parts[2] );
        }

        /**
         * Return licenses filename with full path
         *
         * @param manager LicenseManager The license manager
         * @param name string File name
         * @param testifexits boolean Test if file exists
         * @return string
         */
        public static function getFilenameWithFullPath( $manager, $name, $testifexits = true )
        {
            if( is_null( $manager ) )
            {
                throw new KLicenseException( 'license-error-not-found', self::EIO );
            }
            if( is_null( $name ) )
            {
                throw new KLicenseException( 'license-error-not-found', self::EIO );
            }
            if( PHP_OS == 'WINNT' )
            {
                $_filename = "{$manager->getLicensesDirectory()}\\{$name}";
            }
            else
            {
                $_filename = "{$manager->getLicensesDirectory()}/{$name}";
            }
            if( $testifexits && !is_file( $_filename ) )
            {
                throw new KLicenseException( 'license-error-not-found', self::EIO );
            }
            return $_filename;
        }

        /**
         * Get issued object
         *
         * @return DateTime
         */
        public function getIssued( )
        {
            return $this->issued;
        }

        /**
         * Get expires object
         *
         * @return DateTime
         */
        public function getExpires( )
        {
            return $this->expires;
        }

        /**
         * Check if expired
         *
         * @return boolean
         */
        public function isExpired( )
        {
            return $this->checkIfExpiresIn( $this->manager->getNow( ) );
        }

        /**
         * Check if will expire soon
         *
         * @return boolean
         */
        public function willExpire( )
        {
            return $this->checkIfExpiresIn( $this->manager->getWillExpire( ) );
        }

        /**
         * Check if the license will be expired in supplied DateTime object
         *
         * @param date DateTime date to be compared to
         * @return boolean
         */
        public function checkIfExpiresIn( DateTime $date )
        {
            return $this->expires < $date;
        }

        /**
         * Get the file name
         *
         * @return string
         */
        public function getFilename( )
        {
            return $this->filename;
        }

        /**
         * Magic method to read license values as $obj->keyname
         *
         * @param string name License key
         * @return string
         */
        public function __get( $name )
        {
            return $this->get( $name );
        }

        /**
         * Get license key value
         *
         * @param string name Key
         * @return string
         */
        public function get( $name )
        {
            if( $this->hasKey( $name ) )
            {
                return $this->license[$name];
            }
            throw new KLicenseException( 'license-error-invalid-key', self::EINVALID );
        }

        /**
         * Check if license has key
         *
         * @param string name Key
         * @return boolean
         */
        public function hasKey( $name )
        {
            return array_key_exists( $name, $this->license );
        }

        /**
         * Get existing license keys
         *
         * @return array
         */
        public function getKeys( )
        {
            return array_keys( $this->license );
        }

        /**
         * Get type as string
         *
         * Formatting options:
         *   %n - Type name
         *   %v - Type key value
         *
         * @param glue string types glue
         * @param format string each element formatting settings
         * @return string
         */
        public function getTypeAsString( $glue = ', ', $fmt = '%n' )
        {
            $types = array( );
            if( $this->hasType( self::TYPE_FAX ) )
            {
                $types[] = $this->formatTypeAsString( $fmt, 'fax', $this->fax_channels );
            }
            if( $this->hasType( self::TYPE_ISUP ) )
            {
                $types[] = $this->formatTypeAsString( $fmt, 'isup', $this->isup_circuits );
            }
            if( $this->hasType( self::TYPE_MTP2 ) )
            {
                $types[] = $this->formatTypeAsString( $fmt, 'mtp2', $this->mtp2_links );
            }
            if( $this->hasType( self::TYPE_MTP ) )
            {
                $types[] = $this->formatTypeAsString( $fmt, 'mtp', $this->mtp_dual_mode );
            }
            if( $this->hasType( self::TYPE_G729 ) )
            {
                $types[] = $this->formatTypeAsString( $fmt, 'g729', $this->g729_channels );
            }
            if( $this->hasType( self::TYPE_VOIP ) )
            {
                $types[] = $this->formatTypeAsString( $fmt, 'voip', $this->voip_channels );
            }
            return implode( $glue, $types );
        }

        /**
         * Format type as string
         *
         * @param format string
         * @param name string type name as string
         * @param value string type key value
         * @return string
         */
        protected function formatTypeAsString( $fmt, $name, $value )
        {
            return str_replace( '%v', $value, str_replace( '%n', $name, $fmt ) );
        }

        /**
         * Check if the license is has 'needle' type
         *
         * @param int needle
         * @return boolean
         */
        public function hasType( $needle )
        {
            return ($this->type & $needle) == $needle;
        }

    }
